# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_bincount(backends):
    inputs = [1, 2, 3, 3, 5]
    for B in backends:
        t = B.tensor(inputs)
        cnt = B.bincount(t, minlength=6)
        assert cnt.shape[0] == 6
        assert B.tensor_equal(cnt, B.tensor([0, 1, 1, 2, 0, 1]))
